/*! @file
  @brief
  mruby/c firmware for PSoC5LP(CY8CKIT-059).

  <pre>
  Copyright (C) 2018 Kyushu Institute of Technology.
  Copyright (C) 2018 Shimane IT Open-Innovation Center.

  This file is distributed under BSD 3-Clause License.

  </pre>
*/


#include <project.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include "mrubyc.h"
#include "mrbc_firm.h"


#define MEMORY_SIZE (1024*30)	// mruby/c workarea.
static uint8_t memory_pool[MEMORY_SIZE];

/* EEPROM storage in work flash
  (STRUCTURE e.g. Max bytecode = 4)
 0000 size1	(2bytes * 4)
 0002 size2
 0004 size3
 0006 size4
 0008 IREP1	(size1 bytes)
 xxxx IREP2	(size2 bytes)
 xxxx IREP3	(size3 bytes)
 xxxx IREP4	(size4 bytes)
*/
const uint8_t CYCODE mruby_bytecode[MRBC_SIZE_IREP_STRAGE] = {
  0xff,0xff, 0,0, 0,0, 0,0,
0x52,0x49,0x54,0x45,0x30,0x30,0x30,0x34,0x91,0x2b,0x00,0x00,0x00,0xd7,0x4d,0x41,
0x54,0x5a,0x30,0x30,0x30,0x30,0x49,0x52,0x45,0x50,0x00,0x00,0x00,0x9e,0x30,0x30,
0x30,0x30,0x00,0x00,0x00,0x96,0x00,0x02,0x00,0x05,0x00,0x00,0x00,0x00,0x00,0x19,
0x00,0xbf,0xff,0x83,0x00,0x40,0x09,0x97,0x01,0x00,0x40,0x01,0x01,0xc0,0x00,0x03,
0x01,0x00,0x00,0xb2,0x01,0x40,0x01,0x19,0x00,0xbf,0xff,0x83,0x00,0x40,0x00,0x97,
0x00,0xc0,0x00,0x03,0x01,0x00,0x00,0x06,0x01,0x00,0x40,0x20,0x01,0xc0,0x00,0x03,
0x01,0x00,0x00,0xb2,0x01,0x40,0x00,0x99,0x00,0xc0,0x00,0x03,0x01,0x00,0x00,0x06,
0x01,0x80,0x40,0x01,0x01,0x00,0x80,0xa0,0x01,0x00,0x00,0x06,0x01,0xc0,0x00,0x03,
0x01,0x00,0xc0,0xa0,0x01,0x00,0x00,0x07,0x01,0x3f,0xf5,0x98,0x01,0x00,0x00,0x05,
0x00,0x00,0x00,0x4a,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x04,0x00,0x02,0x3d,0x3d,
0x00,0x00,0x02,0x73,0x77,0x00,0x00,0x03,0x6c,0x65,0x64,0x00,0x00,0x05,0x73,0x6c,
0x65,0x65,0x70,0x00,0x4c,0x56,0x41,0x52,0x00,0x00,0x00,0x1b,0x00,0x00,0x00,0x01,
0x00,0x09,0x6c,0x65,0x64,0x5f,0x6f,0x6e,0x6f,0x66,0x66,0x00,0x00,0x00,0x01,0x45,
0x4e,0x44,0x00,0x00,0x00,0x00,0x08,
};



//================================================================
/*! タイマー割込ハンドラ
*/
CY_ISR(isr_Tick)
{
  mrbc_tick();
}


//================================================================
/*! HAL
*/
int hal_write(int fd, const void *buf, int nbytes)
{
#if 0
  usbuart_write( buf, nbytes );
  return nbytes;
#else
  CONS_PutArray( buf, nbytes );
  return nbytes;
#endif
}

int hal_flush(int fd)
{
  return 0;
}

int _write(int fd, const void *buf, int nbytes)
{
#if 0
  usbuart_write( buf, nbytes );
  return nbytes;
#else
  CONS_PutArray( buf, nbytes );
  return nbytes;
#endif
}


//================================================================
/*! mruby/c methods
*/
static void c_led(mrb_vm *vm, mrb_value v[], int argc)
{
  LED1_Write( GET_INT_ARG(1) );
}

static void c_sw(mrb_vm *vm, mrb_value v[], int argc)
{
  SET_INT_RETURN( !SW1_Read() );
}

void c_pwm(mrb_vm *vm, mrb_value v[], int argc)
{
  int n = GET_INT_ARG(1);     // index of pwm
  int duty = GET_INT_ARG(2);  // duty ratio

  switch(n){
  case 0:
    PWM_1_Stop();
    PWM_1_WriteCompare1( duty );
    PWM_1_Start();
    break;

  case 1:
    PWM_1_Stop();
    PWM_1_WriteCompare2( duty );
    PWM_1_Start();
    break;

  case 2:
    PWM_2_Stop();
    PWM_2_WriteCompare1( duty );
    PWM_2_Start();
    break;

  case 3:
    PWM_2_Stop();
    PWM_2_WriteCompare2( duty );
    PWM_2_Start();
    break;

  default:
    // nothing
    break;
  }
}

void c_output(mrb_vm *vm, mrb_value v[], int argc)
{
  int pin = GET_INT_ARG(1);
  if( pin < 0 || pin > 7 ) return;
  int value = GET_INT_ARG(2);
  int pin_value = Digital_Pin_Read();
  if( value == 0 ){
    pin_value = pin_value & ~(1 << pin);  //reset
  } else {
    pin_value = pin_value | (value << pin);  // set
  }
  Digital_Pin_Write(pin_value);
}


void c_adc(mrb_vm *vm, mrb_value v[], int argc)
{
  int ret = 0;
  int pin = GET_INT_ARG(1);
  if( pin >= 0 && pin <= 3 ){
    ADC_SAR_StartConvert();
    ret = ADC_SAR_GetResult16(pin);
    ADC_SAR_StopConvert();
  }

  SET_INT_RETURN(ret);
}

void c_tone(mrb_vm *vm, mrb_value v[], int argc)
{
  TONE_Stop();
  int freq = GET_INT_ARG(1);

  if( freq > 0 ){
    int period = 12000000 / freq;
    TONE_WritePeriod(period);
    TONE_WriteCompare(period/2);
    TONE_Start();
  }
}

//================================================================
/*! mruby/c main
*/
void mrubyc(void)
{
  mrbc_define_method(0, mrbc_class_object, "led", c_led);
  mrbc_define_method(0, mrbc_class_object, "sw", c_sw);
  mrbc_define_method(0, mrbc_class_object, "pwm", c_pwm);
  mrbc_define_method(0, mrbc_class_object, "output", c_output);
  mrbc_define_method(0, mrbc_class_object, "adc", c_adc);
  mrbc_define_method(0, mrbc_class_object, "tone", c_tone);

  uint16_t *tbl_bytecode_size = (uint16_t *)mruby_bytecode;
  const uint8_t *bytecode = mruby_bytecode + sizeof(uint16_t) * MRBC_MAX_BYTECODES;

  int i;
  for( i = 0 ; i < MRBC_MAX_BYTECODES ; i++ ) {
    if( tbl_bytecode_size[i] == 0 ) break;
    mrbc_create_task( bytecode, 0 );
    bytecode += tbl_bytecode_size[i];
  }

  mrbc_run();
}



//================================================================
/*! check timeout when start

  @retval  0	to monitor mode
  @retval  1	to mruby/c mode (timeout)
*/

int check_timeout(void)
{
  for( int i = 0; i < 256; i++ ) {
    LED1_Write(((i>>4) | (i>>1)) & 0x01 );	// Blink LED

    usbuart_event();
    if( usbuart_can_read_line() ) {
      return 0;
    }
    CyDelay( 10 );
  }

  return 1;
}



//================================================================
/*! main
*/
int main()
{
  /*
    initialize
  */
  CyGlobalIntEnable;
  isr_Tick_StartEx(isr_Tick);
  USBUART_Start(0, USBUART_5V_OPERATION);
  CONS_Start();
  Em_EEPROM_Start();
  PWM_1_Start();
  PWM_2_Start();
  TONE_Start();
  ADC_SAR_Start();

  usbuart_clear();


  mrbc_init(memory_pool, MEMORY_SIZE);

  console_print("\r\n\x1b(B\x1b)B\x1b[2JStart Program.\n");

  /*
    main
  */
  if( check_timeout() == 0 ) {
    LED1_Write( 1 );
    mode_monitor();
  }

  LED1_Write( 0 );
  mrubyc();
}
